# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Python camera library for the Rasperry-Pi camera module
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    print_function,
    division,
    absolute_import,
    )

# Make Py2's str equivalent to Py3's
str = type('')

import ctypes as ct

from . import mmal, mmalobj as mo
from .exc import (
    PiCameraRuntimeError,
    PiCameraValueError,
    mmal_check,
    )


class PiRenderer(object):
    """
    Wraps :class:`~mmalobj.MMALRenderer` for use by PiCamera.

    The *parent* parameter specifies the :class:`PiCamera` instance that has
    constructed this renderer. The *layer* parameter specifies the layer that
    the renderer will inhabit. Higher numbered layers obscure lower numbered
    layers (unless they are partially transparent). The initial opacity of the
    renderer is specified by the *alpha* parameter (which defaults to 255,
    meaning completely opaque). The *fullscreen* parameter which defaults to
    ``True`` indicates whether the renderer should occupy the entire display.
    Finally, the *window* parameter (which only has meaning when *fullscreen*
    is ``False``) is a four-tuple of ``(x, y, width, height)`` which gives the
    screen coordinates that the renderer should occupy when it isn't
    full-screen.

    This base class isn't directly used by :class:`PiCamera`, but the two
    derivatives defined below, :class:`PiOverlayRenderer` and
    :class:`PiPreviewRenderer`, are used to produce overlays and the camera
    preview respectively.
    """

    def __init__(
            self, parent, layer=0, alpha=255, fullscreen=True, window=None,
            crop=None, rotation=0, vflip=False, hflip=False):
        # Create and enable the renderer component
        self._rotation = 0
        self._vflip = False
        self._hflip = False
        self.renderer = mo.MMALRenderer()
        try:
            self.layer = layer
            self.alpha = alpha
            self.fullscreen = fullscreen
            if window is not None:
                self.window = window
            if crop is not None:
                self.crop = crop
            self.rotation = rotation
            self.vflip = vflip
            self.hflip = hflip
            self.renderer.enable()
        except:
            self.renderer.close()
            raise

    def close(self):
        """
        Finalizes the renderer and deallocates all structures.

        This method is called by the camera prior to destroying the renderer
        (or more precisely, letting it go out of scope to permit the garbage
        collector to destroy it at some future time).
        """
        if self.renderer:
            self.renderer.close()
            self.renderer = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def _get_alpha(self):
        return self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION].alpha
    def _set_alpha(self, value):
        try:
            if not (0 <= value <= 255):
                raise PiCameraValueError(
                    "Invalid alpha value: %d (valid range 0..255)" % value)
        except TypeError:
            raise PiCameraValueError("Invalid alpha value: %s" % value)
        mp = self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION]
        mp.set = mmal.MMAL_DISPLAY_SET_ALPHA
        mp.alpha = value
        self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION] = mp
    alpha = property(_get_alpha, _set_alpha, doc="""\
        Retrieves or sets the opacity of the renderer.

        When queried, the :attr:`alpha` property returns a value between 0 and
        255 indicating the opacity of the renderer, where 0 is completely
        transparent and 255 is completely opaque. The default value is 255. The
        property can be set while recordings or previews are in progress.

        .. note::

            If the renderer is being fed RGBA data (as in partially transparent
            overlays), the alpha property will be ignored.
        """)

    def _get_layer(self):
        return self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION].layer
    def _set_layer(self, value):
        try:
            if not (0 <= value <= 255):
                raise PiCameraValueError(
                    "Invalid layer value: %d (valid range 0..255)" % value)
        except TypeError:
            raise PiCameraValueError("Invalid layer value: %s" % value)
        mp = self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION]
        mp.set = mmal.MMAL_DISPLAY_SET_LAYER
        mp.layer = value
        self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION] = mp
    layer = property(_get_layer, _set_layer, doc="""\
        Retrieves or sets the layer of the renderer.

        The :attr:`layer` property is an integer which controls the layer that
        the renderer occupies. Higher valued layers obscure lower valued layers
        (with 0 being the "bottom" layer). The default value is 2. The property
        can be set while recordings or previews are in progress.
        """)

    def _get_fullscreen(self):
        return self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION].fullscreen.value != mmal.MMAL_FALSE
    def _set_fullscreen(self, value):
        mp = self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION]
        mp.set = mmal.MMAL_DISPLAY_SET_FULLSCREEN
        mp.fullscreen = bool(value)
        self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION] = mp
    fullscreen = property(_get_fullscreen, _set_fullscreen, doc="""\
        Retrieves or sets whether the renderer appears full-screen.

        The :attr:`fullscreen` property is a bool which controls whether the
        renderer takes up the entire display or not. When set to ``False``, the
        :attr:`window` property can be used to control the precise size of the
        renderer display. The property can be set while recordings or previews
        are active.
        """)

    def _get_window(self):
        mp = self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION]
        return (
            mp.dest_rect.x,
            mp.dest_rect.y,
            mp.dest_rect.width,
            mp.dest_rect.height,
            )
    def _set_window(self, value):
        try:
            x, y, w, h = value
        except (TypeError, ValueError) as e:
            raise PiCameraValueError(
                "Invalid window rectangle (x, y, w, h) tuple: %s" % value)
        mp = self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION]
        mp.set = mmal.MMAL_DISPLAY_SET_DEST_RECT
        mp.dest_rect = mmal.MMAL_RECT_T(x, y, w, h)
        self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION] = mp
    window = property(_get_window, _set_window, doc="""\
        Retrieves or sets the size of the renderer.

        When the :attr:`fullscreen` property is set to ``False``, the
        :attr:`window` property specifies the size and position of the renderer
        on the display. The property is a 4-tuple consisting of ``(x, y, width,
        height)``. The property can be set while recordings or previews are
        active.
        """)

    def _get_crop(self):
        mp = self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION]
        return (
            mp.src_rect.x,
            mp.src_rect.y,
            mp.src_rect.width,
            mp.src_rect.height,
            )
    def _set_crop(self, value):
        try:
            x, y, w, h = value
        except (TypeError, ValueError) as e:
            raise PiCameraValueError(
                "Invalid crop rectangle (x, y, w, h) tuple: %s" % value)
        mp = self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION]
        mp.set = mmal.MMAL_DISPLAY_SET_SRC_RECT
        mp.src_rect = mmal.MMAL_RECT_T(x, y, w, h)
        self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION] = mp
    crop = property(_get_crop, _set_crop, doc="""\
        Retrieves or sets the area to read from the source.

        The :attr:`crop` property specifies the rectangular area that the
        renderer will read from the source as a 4-tuple of ``(x, y, width,
        height)``. The special value ``(0, 0, 0, 0)`` (which is also the
        default) means to read entire area of the source. The property can be
        set while recordings or previews are active.

        For example, if the camera's resolution is currently configured as
        1280x720, setting this attribute to ``(160, 160, 640, 400)`` will
        crop the preview to the center 640x400 pixels of the input. Note that
        this property does not affect the size of the output rectangle,
        which is controlled with :attr:`fullscreen` and :attr:`window`.

        .. note::

            This property only affects the renderer; it has no bearing on image
            captures or recordings (unlike the :attr:`~PiCamera.zoom` property
            of the :class:`PiCamera` class).
        """)

    def _get_rotation(self):
        return self._rotation
    def _set_rotation(self, value):
        try:
            value = ((int(value) % 360) // 90) * 90
        except ValueError:
            raise PiCameraValueError("Invalid rotation angle: %s" % value)
        self._set_transform(
                self._get_transform(value, self._vflip, self._hflip))
        self._rotation = value
    rotation = property(_get_rotation, _set_rotation, doc="""\
        Retrieves or sets the current rotation of the renderer.

        When queried, the :attr:`rotation` property returns the rotation
        applied to the renderer. Valid values are 0, 90, 180, and 270.

        When set, the property changes the rotation applied to the renderer's
        output. The property can be set while recordings or previews are
        active. The default is 0.

        .. note::

            This property only affects the renderer; it has no bearing on image
            captures or recordings (unlike the :attr:`~PiCamera.rotation`
            property of the :class:`PiCamera` class).
        """)

    def _get_vflip(self):
        return self._vflip
    def _set_vflip(self, value):
        value = bool(value)
        self._set_transform(
                self._get_transform(self._rotation, value, self._hflip))
        self._vflip = value
    vflip = property(_get_vflip, _set_vflip, doc="""\
        Retrieves or sets whether the renderer's output is vertically flipped.

        When queried, the :attr:`vflip` property returns a boolean indicating
        whether or not the renderer's output is vertically flipped. The
        property can be set while recordings or previews are in progress. The
        default is ``False``.

        .. note::

            This property only affects the renderer; it has no bearing on image
            captures or recordings (unlike the :attr:`~PiCamera.vflip` property
            of the :class:`PiCamera` class).
        """)

    def _get_hflip(self):
        return self._hflip
    def _set_hflip(self, value):
        value = bool(value)
        self._set_transform(
                self._get_transform(self._rotation, self._vflip, value))
        self._hflip = value
    hflip = property(_get_hflip, _set_hflip, doc="""\
        Retrieves or sets whether the renderer's output is horizontally
        flipped.

        When queried, the :attr:`vflip` property returns a boolean indicating
        whether or not the renderer's output is horizontally flipped. The
        property can be set while recordings or previews are in progress. The
        default is ``False``.

        .. note::

            This property only affects the renderer; it has no bearing on image
            captures or recordings (unlike the :attr:`~PiCamera.hflip` property
            of the :class:`PiCamera` class).
        """)

    def _get_transform(self, rotate, vflip, hflip):
        # Use a (horizontally) mirrored transform if one of vflip or hflip is
        # set. If vflip is set, rotate by an extra 180 degrees to make up for
        # the lack of a "true" vertical flip
        mirror = vflip ^ hflip
        if vflip:
            rotate = (rotate + 180) % 360
        return {
            (0,   False): mmal.MMAL_DISPLAY_ROT0,
            (90,  False): mmal.MMAL_DISPLAY_ROT90,
            (180, False): mmal.MMAL_DISPLAY_ROT180,
            (270, False): mmal.MMAL_DISPLAY_ROT270,
            (0,   True):  mmal.MMAL_DISPLAY_MIRROR_ROT0,
            (90,  True):  mmal.MMAL_DISPLAY_MIRROR_ROT90,
            (180, True):  mmal.MMAL_DISPLAY_MIRROR_ROT180,
            (270, True):  mmal.MMAL_DISPLAY_MIRROR_ROT270,
            }[(rotate, mirror)]

    def _set_transform(self, value):
        mp = self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION]
        mp.set = mmal.MMAL_DISPLAY_SET_TRANSFORM
        mp.transform = value
        self.renderer.inputs[0].params[mmal.MMAL_PARAMETER_DISPLAYREGION] = mp


class PiOverlayRenderer(PiRenderer):
    """
    Represents an :class:`~mmalobj.MMALRenderer` with a static source for
    overlays.

    This class descends from :class:`PiRenderer` and adds a static *source* for
    the :class:`~mmalobj.MMALRenderer`. The *source* must be an object that
    supports the :ref:`buffer protocol <bufferobjects>` in one of the supported
    formats.

    The optional *resolution* parameter specifies the size of the *source* as a
    ``(width, height)`` tuple. If this is omitted or ``None`` then the
    resolution is assumed to be the same as the parent camera's current
    :attr:`~PiCamera.resolution`. The optional *format* parameter specifies the
    encoding of the *source*. This can be one of the unencoded formats:
    ``'yuv'``, ``'rgb'``, ``'rgba'``, ``'bgr'``, or ``'bgra'``. If omitted or
    ``None``, *format* will be guessed based on the size of *source* (assuming
    3 bytes for `RGB`_, and 4 bytes for `RGBA`_).

    The length of *source* must take into account that widths are rounded up to
    the nearest multiple of 32, and heights to the nearest multiple of 16.  For
    example, if *resolution* is ``(1280, 720)``, and *format* is ``'rgb'`` then
    *source* must be a buffer with length 1280 x 720 x 3 bytes, or 2,764,800
    bytes (because 1280 is a multiple of 32, and 720 is a multiple of 16 no
    extra rounding is required).  However, if *resolution* is ``(97, 57)``, and
    *format* is ``'rgb'`` then *source* must be a buffer with length 128 x 64 x
    3 bytes, or 24,576 bytes (pixels beyond column 97 and row 57 in the source
    will be ignored).

    The *layer*, *alpha*, *fullscreen*, and *window* parameters are the same
    as in :class:`PiRenderer`.

    .. _RGB: https://en.wikipedia.org/wiki/RGB
    .. _RGBA: https://en.wikipedia.org/wiki/RGBA_color_space

    .. versionchanged:: 1.13
        Added *format* parameter
    """

    SOURCE_BPP = {
        3: 'rgb',
        4: 'rgba',
        }

    SOURCE_ENCODINGS = {
        'yuv':  mmal.MMAL_ENCODING_I420,
        'rgb':  mmal.MMAL_ENCODING_RGB24,
        'rgba': mmal.MMAL_ENCODING_RGBA,
        'bgr':  mmal.MMAL_ENCODING_BGR24,
        'bgra': mmal.MMAL_ENCODING_BGRA,
        }

    def __init__(
            self, parent, source, resolution=None, format=None, layer=0,
            alpha=255, fullscreen=True, window=None, crop=None, rotation=0,
            vflip=False, hflip=False):
        super(PiOverlayRenderer, self).__init__(
            parent, layer, alpha, fullscreen, window, crop,
            rotation, vflip, hflip)

        # Copy format from camera's preview port, then adjust the encoding to
        # RGB888 or RGBA and optionally adjust the resolution and size
        if resolution is not None:
            self.renderer.inputs[0].framesize = resolution
        else:
            self.renderer.inputs[0].framesize = parent.resolution
        self.renderer.inputs[0].framerate = 0
        if format is None:
            source_len = mo.buffer_bytes(source)
            plane_size = self.renderer.inputs[0].framesize.pad()
            plane_len = plane_size.width * plane_size.height
            try:
                format = self.SOURCE_BPP[source_len // plane_len]
            except KeyError:
                raise PiCameraValueError(
                    'unable to determine format from source size')
        try:
            self.renderer.inputs[0].format = self.SOURCE_ENCODINGS[format]
        except KeyError:
            raise PiCameraValueError('unknown format %s' % format)
        self.renderer.inputs[0].commit()
        # The following callback is required to prevent the mmalobj layer
        # automatically passing buffers back to the port
        self.renderer.inputs[0].enable(callback=lambda port, buf: True)
        self.update(source)

    def update(self, source):
        """
        Update the overlay with a new source of data.

        The new *source* buffer must have the same size as the original buffer
        used to create the overlay. There is currently no method for changing
        the size of an existing overlay (remove and recreate the overlay if you
        require this).

        .. note::

            If you repeatedly update an overlay renderer, you must make sure
            that you do so at a rate equal to, or slower than, the camera's
            framerate. Going faster will rapidly starve the renderer's pool of
            buffers leading to a runtime error.
        """
        buf = self.renderer.inputs[0].get_buffer()
        buf.data = source
        self.renderer.inputs[0].send_buffer(buf)


class PiPreviewRenderer(PiRenderer):
    """
    Represents an :class:`~mmalobj.MMALRenderer` which uses the camera's
    preview as a source.

    This class descends from :class:`PiRenderer` and adds an
    :class:`~mmalobj.MMALConnection` to connect the renderer to an MMAL port.
    The *source* parameter specifies the :class:`~mmalobj.MMALPort` to connect
    to the renderer.

    The *layer*, *alpha*, *fullscreen*, and *window* parameters are the same
    as in :class:`PiRenderer`.
    """

    def __init__(
            self, parent, source, resolution=None, layer=2, alpha=255,
            fullscreen=True, window=None, crop=None, rotation=0, vflip=False,
            hflip=False):
        super(PiPreviewRenderer, self).__init__(
            parent, layer, alpha, fullscreen, window, crop,
            rotation, vflip, hflip)
        self._parent = parent
        if resolution is not None:
            resolution = mo.to_resolution(resolution)
            source.framesize = resolution
        self.renderer.inputs[0].connect(source).enable()

    def _get_resolution(self):
        result = self._parent._camera.outputs[self._parent.CAMERA_PREVIEW_PORT].framesize
        if result != self._parent.resolution:
            return result
        else:
            return None
    def _set_resolution(self, value):
        if value is not None:
            value = mo.to_resolution(value)
        if (
                value.width > self._parent.resolution.width or
                value.height > self._parent.resolution.height
                ):
            raise PiCameraValueError(
                'preview resolution cannot exceed camera resolution')
        self.renderer.connection.disable()
        if value is None:
            value = self._parent.resolution
        self._parent._camera.outputs[self._parent.CAMERA_PREVIEW_PORT].framesize = value
        self._parent._camera.outputs[self._parent.CAMERA_PREVIEW_PORT].commit()
        self.renderer.connection.enable()
    resolution = property(_get_resolution, _set_resolution, doc="""\
        Retrieves or sets the resolution of the preview renderer.

        By default, the preview's resolution matches the camera's resolution.
        However, particularly high resolutions (such as the maximum resolution
        of the V2 camera module) can cause issues. In this case, you may wish
        to set a lower resolution for the preview that the camera's resolution.

        When queried, the :attr:`resolution` property returns ``None`` if the
        preview's resolution is derived from the camera's. In this case, changing
        the camera's resolution will also cause the preview's resolution to
        change. Otherwise, it returns the current preview resolution as a
        tuple.

        .. note::

            The preview resolution cannot be greater than the camera's
            resolution (in either access). If you set a preview resolution,
            then change the camera's resolution below the preview's resolution,
            this property will silently revert to ``None``, meaning the
            preview's resolution will follow the camera's resolution.

        When set, the property reconfigures the preview renderer with the new
        resolution.  As a special case, setting the property to ``None`` will
        cause the preview to follow the camera's resolution once more. The
        property can be set while recordings are in progress. The default is
        ``None``.

        .. note::

            This property only affects the renderer; it has no bearing on image
            captures or recordings (unlike the :attr:`~PiCamera.resolution`
            property of the :class:`PiCamera` class).

        .. versionadded:: 1.11
        """)


class PiNullSink(object):
    """
    Implements an :class:`~mmalobj.MMALNullSink` which can be used in place of
    a renderer.

    The *parent* parameter specifies the :class:`PiCamera` instance which
    constructed this :class:`~mmalobj.MMALNullSink`. The *source* parameter
    specifies the :class:`~mmalobj.MMALPort` which the null-sink should connect
    to its input.

    The null-sink can act as a drop-in replacement for :class:`PiRenderer` in
    most cases, but obviously doesn't implement attributes like ``alpha``,
    ``layer``, etc. as it simply dumps any incoming frames. This is also the
    reason that this class doesn't derive from :class:`PiRenderer` like all
    other classes in this module.
    """

    def __init__(self, parent, source):
        self.renderer = mo.MMALNullSink()
        self.renderer.enable()
        self.renderer.inputs[0].connect(source).enable()

    def close(self):
        """
        Finalizes the null-sink and deallocates all structures.

        This method is called by the camera prior to destroying the null-sink
        (or more precisely, letting it go out of scope to permit the garbage
        collector to destroy it at some future time).
        """
        if self.renderer:
            self.renderer.close()
            self.renderer = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()





# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Python camera library for the Rasperry-Pi camera module
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    print_function,
    division,
    absolute_import,
    )

# Make Py2's str equivalent to Py3's
str = type('')


import picamera.mmal as mmal


class PiCameraWarning(Warning):
    """
    Base class for PiCamera warnings.
    """


class PiCameraDeprecated(PiCameraWarning, DeprecationWarning):
    """
    Raised when deprecated functionality in picamera is used.
    """


class PiCameraFallback(PiCameraWarning, RuntimeWarning):
    """
    Raised when picamera has to fallback on old functionality.
    """


class PiCameraResizerEncoding(PiCameraWarning, RuntimeWarning):
    """
    Raised when picamera uses a resizer purely for encoding purposes.
    """


class PiCameraAlphaStripping(PiCameraWarning, RuntimeWarning):
    """
    Raised when picamera does alpha-byte stripping.
    """


class PiCameraResolutionRounded(PiCameraWarning, RuntimeWarning):
    """
    Raised when picamera has to round a requested frame size upward.
    """


class PiCameraError(Exception):
    """
    Base class for PiCamera errors.
    """


class PiCameraRuntimeError(PiCameraError, RuntimeError):
    """
    Raised when an invalid sequence of operations is attempted with a
    :class:`PiCamera` object.
    """


class PiCameraClosed(PiCameraRuntimeError):
    """
    Raised when a method is called on a camera which has already been closed.
    """


class PiCameraNotRecording(PiCameraRuntimeError):
    """
    Raised when :meth:`~PiCamera.stop_recording` or
    :meth:`~PiCamera.split_recording` are called against a port which has no
    recording active.
    """


class PiCameraAlreadyRecording(PiCameraRuntimeError):
    """
    Raised when :meth:`~PiCamera.start_recording` or
    :meth:`~PiCamera.record_sequence` are called against a port which already
    has an active recording.
    """


class PiCameraValueError(PiCameraError, ValueError):
    """
    Raised when an invalid value is fed to a :class:`~PiCamera` object.
    """


class PiCameraIOError(PiCameraError, IOError):
    """
    Raised when a :class:`~PiCamera` object is unable to perform an IO
    operation.
    """


class PiCameraMMALError(PiCameraError):
    """
    Raised when an MMAL operation fails for whatever reason.
    """
    def __init__(self, status, prefix=""):
        self.status = status
        PiCameraError.__init__(self, "%s%s%s" % (prefix, ": " if prefix else "", {
            mmal.MMAL_ENOMEM:    "Out of memory",
            mmal.MMAL_ENOSPC:    "Out of resources",
            mmal.MMAL_EINVAL:    "Argument is invalid",
            mmal.MMAL_ENOSYS:    "Function not implemented",
            mmal.MMAL_ENOENT:    "No such file or directory",
            mmal.MMAL_ENXIO:     "No such device or address",
            mmal.MMAL_EIO:       "I/O error",
            mmal.MMAL_ESPIPE:    "Illegal seek",
            mmal.MMAL_ECORRUPT:  "Data is corrupt #FIXME not POSIX",
            mmal.MMAL_ENOTREADY: "Component is not ready #FIXME not POSIX",
            mmal.MMAL_ECONFIG:   "Component is not configured #FIXME not POSIX",
            mmal.MMAL_EISCONN:   "Port is already connected",
            mmal.MMAL_ENOTCONN:  "Port is disconnected",
            mmal.MMAL_EAGAIN:    "Resource temporarily unavailable; try again later",
            mmal.MMAL_EFAULT:    "Bad address",
            }.get(status, "Unknown status error")))


class PiCameraPortDisabled(PiCameraMMALError):
    """
    Raised when attempting a buffer operation on a disabled port.

    This exception is intended for the common use-case of attempting to get
    or send a buffer just when a component is shutting down (e.g. at script
    teardown) and simplifies the trivial response (ignore the error and shut
    down quietly). For example::

        def _callback(self, port, buf):
            try:
                buf = self.outputs[0].get_buffer(False)
            except PiCameraPortDisabled:
                return True # shutting down
            # ...
    """
    def __init__(self, msg):
        super(PiCameraPortDisabled, self).__init__(mmal.MMAL_EINVAL, msg)


def mmal_check(status, prefix=""):
    """
    Checks the return status of an mmal call and raises an exception on
    failure.

    The *status* parameter is the result of an MMAL call. If *status* is
    anything other than MMAL_SUCCESS, a :exc:`PiCameraMMALError` exception is
    raised. The optional *prefix* parameter specifies a prefix message to place
    at the start of the exception's message to provide some context.
    """
    if status != mmal.MMAL_SUCCESS:
        raise PiCameraMMALError(status, prefix)




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Python camera library for the Rasperry-Pi camera module
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    print_function,
    division,
    absolute_import,
    )

# Make Py2's str and range equivalent to Py3's
str = type('')

import warnings
from collections import namedtuple

from picamera.exc import (
    mmal_check,
    PiCameraError,
    PiCameraMMALError,
    PiCameraValueError,
    PiCameraRuntimeError,
    PiCameraDeprecated,
    )


class PiVideoFrameType(object):
    """
    This class simply defines constants used to represent the type of a frame
    in :attr:`PiVideoFrame.frame_type`. Effectively it is a namespace for an
    enum.

    .. attribute:: frame

        Indicates a predicted frame (P-frame). This is the most common frame
        type.

    .. attribute:: key_frame

        Indicates an intra-frame (I-frame) also known as a key frame.

    .. attribute:: sps_header

        Indicates an inline SPS/PPS header (rather than picture data) which is
        typically used as a split point.

    .. attribute:: motion_data

        Indicates the frame is inline motion vector data, rather than picture
        data.

    .. versionadded:: 1.5
    """
    frame = 0
    key_frame = 1
    sps_header = 2
    motion_data = 3


class PiVideoFrame(namedtuple('PiVideoFrame', (
    'index',         # the frame number, where the first frame is 0
    'frame_type',    # a constant indicating the frame type (see PiVideoFrameType)
    'frame_size',    # the size (in bytes) of the frame's data
    'video_size',    # the size (in bytes) of the video so far
    'split_size',    # the size (in bytes) of the video since the last split
    'timestamp',     # the presentation timestamp (PTS) of the frame
    'complete',      # whether the frame is complete or not
    ))):
    """
    This class is a :func:`~collections.namedtuple` derivative used to store
    information about a video frame. It is recommended that you access the
    information stored by this class by attribute name rather than position
    (for example: ``frame.index`` rather than ``frame[0]``).

    .. attribute:: index

        Returns the zero-based number of the frame. This is a monotonic counter
        that is simply incremented every time the camera starts outputting a
        new frame. As a consequence, this attribute cannot be used to detect
        dropped frames. Nor does it necessarily represent actual frames; it
        will be incremented for SPS headers and motion data buffers too.

    .. attribute:: frame_type

        Returns a constant indicating the kind of data that the frame contains
        (see :class:`PiVideoFrameType`). Please note that certain frame types
        contain no image data at all.

    .. attribute:: frame_size

        Returns the size in bytes of the current frame. If a frame is written
        in multiple chunks, this value will increment while :attr:`index`
        remains static. Query :attr:`complete` to determine whether the frame
        has been completely output yet.

    .. attribute:: video_size

        Returns the size in bytes of the entire video up to this frame.  Note
        that this is unlikely to match the size of the actual file/stream
        written so far. This is because a stream may utilize buffering which
        will cause the actual amount written (e.g. to disk) to lag behind the
        value reported by this attribute.

    .. attribute:: split_size

        Returns the size in bytes of the video recorded since the last call to
        either :meth:`~PiCamera.start_recording` or
        :meth:`~PiCamera.split_recording`. For the reasons explained above,
        this may differ from the size of the actual file/stream written so far.

    .. attribute:: timestamp

        Returns the presentation timestamp (PTS) of the frame. This represents
        the point in time that the Pi received the first line of the frame from
        the camera.

        The timestamp is measured in microseconds (millionths of a second).
        When the camera's clock mode is ``'reset'`` (the default), the
        timestamp is relative to the start of the video recording.  When the
        camera's :attr:`~PiCamera.clock_mode` is ``'raw'``, it is relative to
        the last system reboot. See :attr:`~PiCamera.timestamp` for more
        information.

        .. warning::

            Currently, the camera occasionally returns "time unknown" values in
            this field which picamera represents as ``None``. If you are
            querying this property you will need to check the value is not
            ``None`` before using it. This happens for SPS header "frames",
            for example.

    .. attribute:: complete

        Returns a bool indicating whether the current frame is complete or not.
        If the frame is complete then :attr:`frame_size` will not increment
        any further, and will reset for the next frame.

    .. versionchanged:: 1.5
        Deprecated :attr:`header` and :attr:`keyframe` attributes and added the
        new :attr:`frame_type` attribute instead.

    .. versionchanged:: 1.9
        Added the :attr:`complete` attribute.
    """

    __slots__ = () # workaround python issue #24931

    @property
    def position(self):
        """
        Returns the zero-based position of the frame in the stream containing
        it.
        """
        return self.split_size - self.frame_size

    @property
    def keyframe(self):
        """
        Returns a bool indicating whether the current frame is a keyframe (an
        intra-frame, or I-frame in MPEG parlance).

        .. deprecated:: 1.5
            Please compare :attr:`frame_type` to
            :attr:`PiVideoFrameType.key_frame` instead.
        """
        warnings.warn(
            PiCameraDeprecated(
                'PiVideoFrame.keyframe is deprecated; please check '
                'PiVideoFrame.frame_type for equality with '
                'PiVideoFrameType.key_frame instead'))
        return self.frame_type == PiVideoFrameType.key_frame

    @property
    def header(self):
        """
        Contains a bool indicating whether the current frame is actually an
        SPS/PPS header. Typically it is best to split an H.264 stream so that
        it starts with an SPS/PPS header.

        .. deprecated:: 1.5
            Please compare :attr:`frame_type` to
            :attr:`PiVideoFrameType.sps_header` instead.
        """
        warnings.warn(
            PiCameraDeprecated(
                'PiVideoFrame.header is deprecated; please check '
                'PiVideoFrame.frame_type for equality with '
                'PiVideoFrameType.sps_header instead'))
        return self.frame_type == PiVideoFrameType.sps_header




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Python camera library for the Rasperry-Pi camera module
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    print_function,
    division,
    absolute_import,
    )
# Make Py2's zip equivalent to Py3's
try:
    from itertools import izip as zip
except ImportError:
    pass

# Make Py2's str and range equivalent to Py3's
str = type('')


import colorsys
from math import pi, sqrt, atan2, degrees, radians, sin, cos, exp
from fractions import Fraction
from collections import namedtuple


# From the CSS Color Module Level 3 specification, section 4.3
# <http://www.w3.org/TR/css3-color/#svg-color>
NAMED_COLORS = {
    'aliceblue':             '#f0f8ff',
    'antiquewhite':          '#faebd7',
    'aqua':                  '#00ffff',
    'aquamarine':            '#7fffd4',
    'azure':                 '#f0ffff',
    'beige':                 '#f5f5dc',
    'bisque':                '#ffe4c4',
    'black':                 '#000000',
    'blanchedalmond':        '#ffebcd',
    'blue':                  '#0000ff',
    'blueviolet':            '#8a2be2',
    'brown':                 '#a52a2a',
    'burlywood':             '#deb887',
    'cadetblue':             '#5f9ea0',
    'chartreuse':            '#7fff00',
    'chocolate':             '#d2691e',
    'coral':                 '#ff7f50',
    'cornflowerblue':        '#6495ed',
    'cornsilk':              '#fff8dc',
    'crimson':               '#dc143c',
    'cyan':                  '#00ffff',
    'darkblue':              '#00008b',
    'darkcyan':              '#008b8b',
    'darkgoldenrod':         '#b8860b',
    'darkgray':              '#a9a9a9',
    'darkgreen':             '#006400',
    'darkgrey':              '#a9a9a9',
    'darkkhaki':             '#bdb76b',
    'darkmagenta':           '#8b008b',
    'darkolivegreen':        '#556b2f',
    'darkorange':            '#ff8c00',
    'darkorchid':            '#9932cc',
    'darkred':               '#8b0000',
    'darksalmon':            '#e9967a',
    'darkseagreen':          '#8fbc8f',
    'darkslateblue':         '#483d8b',
    'darkslategray':         '#2f4f4f',
    'darkslategrey':         '#2f4f4f',
    'darkturquoise':         '#00ced1',
    'darkviolet':            '#9400d3',
    'deeppink':              '#ff1493',
    'deepskyblue':           '#00bfff',
    'dimgray':               '#696969',
    'dimgrey':               '#696969',
    'dodgerblue':            '#1e90ff',
    'firebrick':             '#b22222',
    'floralwhite':           '#fffaf0',
    'forestgreen':           '#228b22',
    'fuchsia':               '#ff00ff',
    'gainsboro':             '#dcdcdc',
    'ghostwhite':            '#f8f8ff',
    'gold':                  '#ffd700',
    'goldenrod':             '#daa520',
    'gray':                  '#808080',
    'green':                 '#008000',
    'greenyellow':           '#adff2f',
    'grey':                  '#808080',
    'honeydew':              '#f0fff0',
    'hotpink':               '#ff69b4',
    'indianred':             '#cd5c5c',
    'indigo':                '#4b0082',
    'ivory':                 '#fffff0',
    'khaki':                 '#f0e68c',
    'lavender':              '#e6e6fa',
    'lavenderblush':         '#fff0f5',
    'lawngreen':             '#7cfc00',
    'lemonchiffon':          '#fffacd',
    'lightblue':             '#add8e6',
    'lightcoral':            '#f08080',
    'lightcyan':             '#e0ffff',
    'lightgoldenrodyellow':  '#fafad2',
    'lightgray':             '#d3d3d3',
    'lightgreen':            '#90ee90',
    'lightgrey':             '#d3d3d3',
    'lightpink':             '#ffb6c1',
    'lightsalmon':           '#ffa07a',
    'lightseagreen':         '#20b2aa',
    'lightskyblue':          '#87cefa',
    'lightslategray':        '#778899',
    'lightslategrey':        '#778899',
    'lightsteelblue':        '#b0c4de',
    'lightyellow':           '#ffffe0',
    'lime':                  '#00ff00',
    'limegreen':             '#32cd32',
    'linen':                 '#faf0e6',
    'magenta':               '#ff00ff',
    'maroon':                '#800000',
    'mediumaquamarine':      '#66cdaa',
    'mediumblue':            '#0000cd',
    'mediumorchid':          '#ba55d3',
    'mediumpurple':          '#9370db',
    'mediumseagreen':        '#3cb371',
    'mediumslateblue':       '#7b68ee',
    'mediumspringgreen':     '#00fa9a',
    'mediumturquoise':       '#48d1cc',
    'mediumvioletred':       '#c71585',
    'midnightblue':          '#191970',
    'mintcream':             '#f5fffa',
    'mistyrose':             '#ffe4e1',
    'moccasin':              '#ffe4b5',
    'navajowhite':           '#ffdead',
    'navy':                  '#000080',
    'oldlace':               '#fdf5e6',
    'olive':                 '#808000',
    'olivedrab':             '#6b8e23',
    'orange':                '#ffa500',
    'orangered':             '#ff4500',
    'orchid':                '#da70d6',
    'palegoldenrod':         '#eee8aa',
    'palegreen':             '#98fb98',
    'paleturquoise':         '#afeeee',
    'palevioletred':         '#db7093',
    'papayawhip':            '#ffefd5',
    'peachpuff':             '#ffdab9',
    'peru':                  '#cd853f',
    'pink':                  '#ffc0cb',
    'plum':                  '#dda0dd',
    'powderblue':            '#b0e0e6',
    'purple':                '#800080',
    'red':                   '#ff0000',
    'rosybrown':             '#bc8f8f',
    'royalblue':             '#4169e1',
    'saddlebrown':           '#8b4513',
    'salmon':                '#fa8072',
    'sandybrown':            '#f4a460',
    'seagreen':              '#2e8b57',
    'seashell':              '#fff5ee',
    'sienna':                '#a0522d',
    'silver':                '#c0c0c0',
    'skyblue':               '#87ceeb',
    'slateblue':             '#6a5acd',
    'slategray':             '#708090',
    'slategrey':             '#708090',
    'snow':                  '#fffafa',
    'springgreen':           '#00ff7f',
    'steelblue':             '#4682b4',
    'tan':                   '#d2b48c',
    'teal':                  '#008080',
    'thistle':               '#d8bfd8',
    'tomato':                '#ff6347',
    'turquoise':             '#40e0d0',
    'violet':                '#ee82ee',
    'wheat':                 '#f5deb3',
    'white':                 '#ffffff',
    'whitesmoke':            '#f5f5f5',
    'yellow':                '#ffff00',
    'yellowgreen':           '#9acd32',
    }


class Red(float):
    """
    Represents the red component of a :class:`Color` for use in
    transformations. Instances of this class can be constructed directly with a
    float value, or by querying the :attr:`Color.red` attribute. Addition,
    subtraction, and multiplication are supported with :class:`Color`
    instances. For example::

        >>> Color.from_rgb(0, 0, 0) + Red(0.5)
        <Color "#7f0000">
        >>> Color('#f00') - Color('#900').red
        <Color "#660000">
        >>> (Red(0.1) * Color('red')).red
        Red(0.1)
    """

    def __repr__(self):
        return "Red(%s)" % self


class Green(float):
    """
    Represents the green component of a :class:`Color` for use in
    transformations.  Instances of this class can be constructed directly with
    a float value, or by querying the :attr:`Color.green` attribute. Addition,
    subtraction, and multiplication are supported with :class:`Color`
    instances. For example::

        >>> Color(0, 0, 0) + Green(0.1)
        <Color "#001900">
        >>> Color.from_yuv(1, -0.4, -0.6) - Green(1)
        <Color "#50002f">
        >>> (Green(0.5) * Color('white')).rgb
        (Red(1.0), Green(0.5), Blue(1.0))
    """

    def __repr__(self):
        return "Green(%s)" % self


class Blue(float):
    """
    Represents the blue component of a :class:`Color` for use in
    transformations.  Instances of this class can be constructed directly with
    a float value, or by querying the :attr:`Color.blue` attribute. Addition,
    subtraction, and multiplication are supported with :class:`Color`
    instances. For example::

        >>> Color(0, 0, 0) + Blue(0.2)
        <Color "#000033">
        >>> Color.from_hls(0.5, 0.5, 1.0) - Blue(1)
        <Color "#00fe00">
        >>> Blue(0.9) * Color('white')
        <Color "#ffffe5">
    """

    def __repr__(self):
        return "Blue(%s)" % self


class Hue(float):
    """
    Represents the hue of a :class:`Color` for use in transformations.
    Instances of this class can be constructed directly with a float value in
    the range [0.0, 1.0) representing an angle around the `HSL hue wheel`_. As
    this is a circular mapping, 0.0 and 1.0 effectively mean the same thing,
    i.e.  out of range values will be normalized into the range [0.0, 1.0).

    The class can also be constructed with the keyword arguments ``deg`` or
    ``rad`` if you wish to specify the hue value in degrees or radians instead,
    respectively. Instances can also be constructed by querying the
    :attr:`Color.hue` attribute.

    Addition, subtraction, and multiplication are supported with :class:`Color`
    instances. For example::

        >>> Color(1, 0, 0).hls
        (0.0, 0.5, 1.0)
        >>> (Color(1, 0, 0) + Hue(deg=180)).hls
        (0.5, 0.5, 1.0)

    Note that whilst multiplication by a :class:`Hue` doesn't make much sense,
    it is still supported. However, the circular nature of a hue value can lead
    to suprising effects. In particular, since 1.0 is equivalent to 0.0 the
    following may be observed::

        >>> (Hue(1.0) * Color.from_hls(0.5, 0.5, 1.0)).hls
        (0.0, 0.5, 1.0)

    .. _HSL hue wheel: https://en.wikipedia.org/wiki/Hue
    """

    def __new__(cls, n=None, deg=None, rad=None):
        if n is not None:
            return super(Hue, cls).__new__(cls, n % 1.0)
        elif deg is not None:
            return super(Hue, cls).__new__(cls, (deg / 360.0) % 1.0)
        elif rad is not None:
            return super(Hue, cls).__new__(cls, (rad / (2 * pi)) % 1.0)
        else:
            raise ValueError('You must specify a value, or deg or rad')

    def __repr__(self):
        return "Hue(deg=%s)" % self.deg

    @property
    def deg(self):
        return self * 360.0

    @property
    def rad(self):
        return self * 2 * pi


class Lightness(float):
    """
    Represents the lightness of a :class:`Color` for use in transformations.
    Instances of this class can be constructed directly with a float value, or
    by querying the :attr:`Color.lightness` attribute. Addition, subtraction,
    and multiplication are supported with :class:`Color` instances. For
    example::

        >>> Color(0, 0, 0) + Lightness(0.1)
        <Color "#191919">
        >>> Color.from_rgb_bytes(0x80, 0x80, 0) - Lightness(0.2)
        <Color "#191900">
        >>> Lightness(0.9) * Color('wheat')
        <Color "#f0cd8d">
    """

    def __repr__(self):
        return "Lightness(%s)" % self


class Saturation(float):
    """
    Represents the saturation of a :class:`Color` for use in transformations.
    Instances of this class can be constructed directly with a float value, or
    by querying the :attr:`Color.saturation` attribute. Addition, subtraction,
    and multiplication are supported with :class:`Color` instances. For
    example::

        >>> Color(0.9, 0.9, 0.6) + Saturation(0.1)
        <Color "#ebeb92">
        >>> Color('red') - Saturation(1)
        <Color "#7f7f7f">
        >>> Saturation(0.5) * Color('wheat')
        <Color "#e4d9c3">
    """

    def __repr__(self):
        return "Lightness(%s)" % self



clamp_float = lambda v: max(0.0, min(1.0, v))
clamp_bytes = lambda v: max(0, min(255, v))
to_srgb = lambda c: 12.92 * c if c <= 0.0031308 else (1.055 * c ** (1/2.4) - 0.055)
from_srgb = lambda c: c / 12.92 if c <= 0.04045 else ((c + 0.055) / 1.055) ** 2.4
D65 = (0.95047, 1.0, 1.08883)
U = lambda x, y, z: 4 * x / (x + 15 * y + 3 * z)
V = lambda x, y, z: 9 * y / (x + 15 * y + 3 * z)
matrix_mult = lambda m, n: (
        sum(mval * nval for mval, nval in zip(mrow, n))
        for mrow in m
        )

class Color(namedtuple('Color', ('red', 'green', 'blue'))):
    """
    The Color class is a tuple which represents a color as red, green, and
    blue components.

    The class has a flexible constructor which allows you to create an instance
    from a variety of color systems including `RGB`_, `Y'UV`_, `Y'IQ`_, `HLS`_,
    and `HSV`_.  There are also explicit constructors for each of these systems
    to allow you to force the use of a system in your code. For example, an
    instance of :class:`Color` can be constructed in any of the following
    ways::

        >>> Color('#f00')
        <Color "#ff0000">
        >>> Color('green')
        <Color "#008000">
        >>> Color(0, 0, 1)
        <Color "#0000ff">
        >>> Color(hue=0, saturation=1, value=0.5)
        <Color "#7f0000">
        >>> Color(y=0.4, u=-0.05, v=0.615)
        <Color "#ff0f4c">

    The specific forms that the default constructor will accept are enumerated
    below:

    .. tabularcolumns:: |p{40mm}|p{100mm}|

    +------------------------------+------------------------------------------+
    | Style                        | Description                              |
    +==============================+==========================================+
    | Single positional parameter  | Equivalent to calling                    |
    |                              | :meth:`Color.from_string`.               |
    +------------------------------+------------------------------------------+
    | Three positional parameters  | Equivalent to calling                    |
    |                              | :meth:`Color.from_rgb` if all three      |
    |                              | parameters are between 0.0 and 1.0, or   |
    |                              | :meth:`Color.from_rgb_bytes` otherwise.  |
    +------------------------------+                                          |
    | Three named parameters:      |                                          |
    | *r*, *g*, *b*                |                                          |
    +------------------------------+                                          |
    | Three named parameters:      |                                          |
    | *red*, *green*, *blue*       |                                          |
    +------------------------------+------------------------------------------+
    | Three named parameters:      | Equivalent to calling                    |
    | *y*, *u*, *v*                | :meth:`Color.from_yuv` if *y* is between |
    |                              | 0.0 and 1.0, *u* is between -0.436 and   |
    |                              | 0.436, and *v* is between -0.615 and     |
    |                              | 0.615, or :meth:`Color.from_yuv_bytes`   |
    |                              | otherwise.                               |
    +------------------------------+------------------------------------------+
    | Three named parameters:      | Equivalent to calling                    |
    | *y*, *i*, *q*                | :meth:`Color.from_yiq`.                  |
    +------------------------------+------------------------------------------+
    | Three named parameters:      | Equivalent to calling                    |
    | *h*, *l*, *s*                | :meth:`Color.from_hls`.                  |
    +------------------------------+                                          |
    | Three named parameters:      |                                          |
    | *hue*, *lightness*,          |                                          |
    | *saturation*                 |                                          |
    +------------------------------+------------------------------------------+
    | Three named parameters:      | Equivalent to calling                    |
    | *h*, *s*, *v*                | :meth:`Color.from_hsv`                   |
    +------------------------------+                                          |
    | Three named parameters:      |                                          |
    | *hue*, *saturation*, *value* |                                          |
    +------------------------------+------------------------------------------+
    | Three named parameters:      | Equivalent to calling                    |
    | *x*, *y*, *z*                | :meth:`Color.from_cie_xyz`               |
    +------------------------------+------------------------------------------+
    | Three named parameters:      | Equivalent to calling                    |
    | *l*, *a*, *b*                | :meth:`Color.from_cie_lab`               |
    +------------------------------+------------------------------------------+
    | Three named parameters:      | Equivalent to calling                    |
    | *l*, *u*, *v*                | :meth:`Color.from_cie_luv`               |
    +------------------------------+------------------------------------------+

    If the constructor parameters do not conform to any of the variants in the
    table above, a :exc:`ValueError` will be thrown.

    Internally, the color is *always* represented as 3 float values
    corresponding to the red, green, and blue components of the color. These
    values take a value from 0.0 to 1.0 (least to full intensity). The class
    provides several attributes which can be used to convert one color system
    into another::

        >>> Color('#f00').hls
        (0.0, 0.5, 1.0)
        >>> Color.from_string('green').hue
        Hue(deg=120.0)
        >>> Color.from_rgb_bytes(0, 0, 255).yuv
        (0.114, 0.435912, -0.099978)

    As :class:`Color` derives from tuple, instances are immutable. While this
    provides the advantage that they can be used as keys in a dict, it does
    mean that colors themselves cannot be directly manipulated (e.g. by
    reducing the red component).

    However, several auxilliary classes in the module provide the ability to
    perform simple transformations of colors via operators which produce a new
    :class:`Color` instance. For example::

        >>> Color('red') - Red(0.5)
        <Color "#7f0000">
        >>> Color('green') + Red(0.5)
        <Color "#7f8000">
        >>> Color.from_hls(0.5, 0.5, 1.0)
        <Color "#00feff">
        >>> Color.from_hls(0.5, 0.5, 1.0) * Lightness(0.8)
        <Color "#00cbcc">
        >>> (Color.from_hls(0.5, 0.5, 1.0) * Lightness(0.8)).hls
        (0.5, 0.4, 1.0)

    From the last example above one can see that even attributes not directly
    stored by the color (such as lightness) can be manipulated in this fashion.
    In this case a :class:`Color` instance is constructed from HLS (hue,
    lightness, saturation) values with a lightness of 0.5. This is multiplied
    by a :class:`Lightness` instance with a value of 0.8 which constructs a new
    :class:`Color` with the same hue and saturation, but a lightness of 0.5 *
    0.8 = 0.4.

    If an instance is converted to a string (with :func:`str`) it will return a
    string containing the 7-character HTML code for the color (e.g. "#ff0000"
    for red). As can be seen in the examples above, a similar representation is
    returned for :func:`repr`.

    .. _RGB: https://en.wikipedia.org/wiki/RGB_color_space
    .. _Y'UV: https://en.wikipedia.org/wiki/YUV
    .. _Y'IQ: https://en.wikipedia.org/wiki/YIQ
    .. _HLS: https://en.wikipedia.org/wiki/HSL_and_HSV
    .. _HSV: https://en.wikipedia.org/wiki/HSL_and_HSV
    """

    def __new__(cls, *args, **kwargs):
        def from_rgb(r, g, b):
            if 0.0 <= r <= 1.0 and 0.0 <= g <= 1.0 and 0.0 <= b <= 1.0:
                return cls.from_rgb(r, g, b)
            else:
                return cls.from_rgb_bytes(r, g, b)

        def from_yuv(y, u, v):
            if 0.0 <= y <= 1.0 and -0.436 <= u <= 0.436 and -0.615 <= v <= 0.615:
                return cls.from_yuv(y, u, v)
            else:
                return cls.from_yuv_bytes(y, u, v)

        if kwargs:
            try:
                return {
                    frozenset('rgb'):   from_rgb,
                    frozenset('yuv'):   from_yuv,
                    frozenset('yiq'):   cls.from_yiq,
                    frozenset('hls'):   cls.from_hls,
                    frozenset('hsv'):   cls.from_hsv,
                    frozenset('xyz'):   cls.from_cie_xyz,
                    frozenset('lab'):   cls.from_cie_lab,
                    frozenset('luv'):   cls.from_cie_luv,
                    frozenset(('red', 'green', 'blue')):
                        lambda red, green, blue: from_rgb(red, green, blue),
                    frozenset(('hue', 'lightness', 'saturation')):
                        lambda hue, lightness, saturation: cls.from_hls(hue, lightness, saturation),
                    frozenset(('hue', 'saturation', 'value')):
                        lambda hue, saturation, value: cls.from_hsv(hue, saturation, value),
                    }[frozenset(kwargs.keys())](**kwargs)
            except KeyError:
                pass
        else:
            if len(args) == 1:
                return cls.from_string(args[0])
            elif len(args) == 3:
                return from_rgb(*args)
        raise ValueError('Unable to construct Color from provided arguments')

    @classmethod
    def from_string(cls, s):
        """
        Construct a :class:`Color` from a 4 or 7 character CSS-like
        representation (e.g. "#f00" or "#ff0000" for red), or from one of the
        named colors (e.g. "green" or "wheat") from the `CSS standard`_. Any
        other string format will result in a :exc:`ValueError`.

        .. _CSS standard: http://www.w3.org/TR/css3-color/#svg-color
        """
        if isinstance(s, bytes):
            s = s.decode('ascii')
        if s.startswith('#'):
            if len(s) == 7:
                return cls.from_rgb_bytes(
                    int(s[1:3], base=16),
                    int(s[3:5], base=16),
                    int(s[5:7], base=16)
                    )
            elif len(s) == 4:
                return cls.from_rgb_bytes(
                    int(s[1:2], base=16) * 0x11,
                    int(s[2:3], base=16) * 0x11,
                    int(s[3:4], base=16) * 0x11
                    )
            raise ValueError('Unrecognized color format "%s"' % s)
        try:
            return cls.from_string(NAMED_COLORS[s.lower()])
        except KeyError:
            raise ValueError('Unrecognized color name "%s"' % s)

    @classmethod
    def from_rgb(cls, r, g, b):
        """
        Construct a :class:`Color` from three `RGB`_ float values between 0.0
        and 1.0.
        """
        return super(Color, cls).__new__(cls, r, g, b)

    @classmethod
    def from_rgb_565(cls, n):
        """
        Construct a :class:`Color` from an unsigned 16-bit integer number
        in RGB565 format.
        """
        r = (n & 0xF800) / 0xF800
        g = (n & 0x07E0) / 0x07E0
        b = (n & 0x001F) / 0x001F
        return super(Color, cls).__new__(cls, r, g, b)

    @classmethod
    def from_rgb_bytes(cls, r, g, b):
        """
        Construct a :class:`Color` from three `RGB`_ byte values between 0 and
        255.
        """
        return super(Color, cls).__new__(cls, r / 255.0, g / 255.0, b / 255.0)

    @classmethod
    def from_yuv(cls, y, u, v):
        """
        Construct a :class:`Color` from three `Y'UV`_ float values. The Y value
        may be between 0.0 and 1.0. U may be between -0.436 and 0.436, while
        V may be between -0.615 and 0.615.
        """
        return super(Color, cls).__new__(
                cls,
                clamp_float(y + 1.14  * v),
                clamp_float(y - 0.395 * u - 0.581 * v),
                clamp_float(y + 2.033 * u),
                )

    @classmethod
    def from_yuv_bytes(cls, y, u, v):
        """
        Construct a :class:`Color` from three `Y'UV`_ byte values between 0 and
        255. The U and V values are biased by 128 to prevent negative values as
        is typical in video applications. The Y value is biased by 16 for the
        same purpose.
        """
        c = y - 16
        d = u - 128
        e = v - 128
        return cls.from_rgb_bytes(
                clamp_bytes((298 * c + 409 * e + 128) >> 8),
                clamp_bytes((298 * c - 100 * d - 208 * e + 128) >> 8),
                clamp_bytes((298 * c + 516 * d + 128) >> 8),
                )

    @classmethod
    def from_yiq(cls, y, i, q):
        """
        Construct a :class:`Color` from three `Y'IQ`_ float values. Y' can be
        between 0.0 and 1.0, while I and Q can be between -1.0 and 1.0.
        """
        return super(Color, cls).__new__(cls, *colorsys.yiq_to_rgb(y, i, q))

    @classmethod
    def from_hls(cls, h, l, s):
        """
        Construct a :class:`Color` from `HLS`_ (hue, lightness, saturation)
        floats between 0.0 and 1.0.
        """
        return super(Color, cls).__new__(cls, *colorsys.hls_to_rgb(h, l, s))

    @classmethod
    def from_hsv(cls, h, s, v):
        """
        Construct a :class:`Color` from `HSV`_ (hue, saturation, value) floats
        between 0.0 and 1.0.
        """
        return super(Color, cls).__new__(cls, *colorsys.hsv_to_rgb(h, s, v))

    @classmethod
    def from_cie_xyz(cls, x, y, z):
        """
        Construct a :class:`Color` from (X, Y, Z) float values representing
        a color in the `CIE 1931 color space`_. The conversion assumes the
        sRGB working space with reference white D65.

        .. _CIE 1931 color space: https://en.wikipedia.org/wiki/CIE_1931_color_space
        """
        m = matrix_mult(
            (( 3.2404542, -1.5371385, -0.4985314),
             (-0.9692660,  1.8760108,  0.0415560),
             ( 0.0556434, -0.2040259,  1.0572252),
             ),
            (x,
             y,
             z,
             ))
        return super(Color, cls).__new__(cls, *(to_srgb(c) for c in m))

    @classmethod
    def from_cie_lab(cls, l, a, b):
        """
        Construct a :class:`Color` from (L*, a*, b*) float values representing
        a color in the `CIE Lab color space`_. The conversion assumes the
        sRGB working space with reference white D65.

        .. _CIE Lab color space: https://en.wikipedia.org/wiki/Lab_color_space
        """
        theta = Fraction(6, 29)
        fy = (l + 16) / 116
        fx = fy + a / 500
        fz = fy - b / 200
        xyz = (
            n ** 3 if n > theta else 3 * theta ** 2 * (n - Fraction(4, 29))
            for n in (fx, fy, fz)
            )
        return cls.from_cie_xyz(*(n * m for n, m in zip(xyz, D65)))

    @classmethod
    def from_cie_luv(cls, l, u, v):
        """
        Construct a :class:`Color` from (L*, u*, v*) float values representing
        a color in the `CIE Luv color space`_. The conversion assumes the sRGB
        working space with reference white D65.

        .. _CIE Luv color space: https://en.wikipedia.org/wiki/CIELUV
        """
        uw = U(*D65)
        vw = V(*D65)
        u_p = u / (13 * l) + uw
        v_p = v / (13 * l) + vw
        y = D65[1] * (l * Fraction(3, 29) ** 3 if l <= 8 else ((l + 16) / 116) ** 3)
        x = y * (9 * u_p) / (4 * v_p)
        z = y * (12 - 3 * u_p - 20 * v_p) / (4 * v_p)
        return cls.from_cie_xyz(x, y, z)

    def __add__(self, other):
        if isinstance(other, Red):
            return Color(clamp_float(self.red + other), self.green, self.blue)
        elif isinstance(other, Green):
            return Color(self.red, clamp_float(self.green + other), self.blue)
        elif isinstance(other, Blue):
            return Color(self.red, self.green, clamp_float(self.blue + other))
        elif isinstance(other, Hue):
            h, l, s = self.hls
            return Color.from_hls((h + other) % 1.0, l, s)
        elif isinstance(other, Lightness):
            h, l, s = self.hls
            return Color.from_hls(h, clamp_float(l + other), s)
        elif isinstance(other, Saturation):
            h, l, s = self.hls
            return Color.from_hls(h, l, clamp_float(s + other))
        return NotImplemented

    def __radd__(self, other):
        # Addition is commutative
        if isinstance(other, (Red, Green, Blue, Hue, Lightness, Saturation)):
            return self.__add__(other)
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Red):
            return Color(clamp_float(self.red - other), self.green, self.blue)
        elif isinstance(other, Green):
            return Color(self.red, clamp_float(self.green - other), self.blue)
        elif isinstance(other, Blue):
            return Color(self.red, self.green, clamp_float(self.blue - other))
        elif isinstance(other, Hue):
            h, l, s = self.hls
            return Color.from_hls((h - other) % 1.0, l, s)
        elif isinstance(other, Lightness):
            h, l, s = self.hls
            return Color.from_hls(h, clamp_float(l - other), s)
        elif isinstance(other, Saturation):
            h, l, s = self.hls
            return Color.from_hls(h, l, clamp_float(s - other))
        return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, Red):
            return Color(clamp_float(other - self.red), self.green, self.blue)
        elif isinstance(other, Green):
            return Color(self.red, clamp_float(other - self.green), self.blue)
        elif isinstance(other, Blue):
            return Color(self.red, self.green, clamp_float(other - self.blue))
        elif isinstance(other, Hue):
            h, l, s = self.hls
            return Color.from_hls((other - h) % 1.0, l, s)
        elif isinstance(other, Lightness):
            h, l, s = self.hls
            return Color.from_hls(h, clamp_float(other - l), s)
        elif isinstance(other, Saturation):
            h, l, s = self.hls
            return Color.from_hls(h, l, clamp_float(other - s))
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, Red):
            return Color(clamp_float(self.red * other), self.green, self.blue)
        elif isinstance(other, Green):
            return Color(self.red, clamp_float(self.green * other), self.blue)
        elif isinstance(other, Blue):
            return Color(self.red, self.green, clamp_float(self.blue * other))
        elif isinstance(other, Hue):
            h, l, s = self.hls
            return Color.from_hls((h * other) % 1.0, l, s)
        elif isinstance(other, Lightness):
            h, l, s = self.hls
            return Color.from_hls(h, clamp_float(l * other), s)
        elif isinstance(other, Saturation):
            h, l, s = self.hls
            return Color.from_hls(h, l, clamp_float(s * other))
        return NotImplemented

    def __rmul__(self, other):
        # Multiplication is commutative
        if isinstance(other, (Red, Green, Blue, Hue, Lightness, Saturation)):
            return self.__mul__(other)

    def __str__(self):
        return '#%02x%02x%02x' % self.rgb_bytes

    def __repr__(self):
        return '<Color "%s">' % str(self)

    @property
    def rgb(self):
        """
        Returns a 3-tuple of (red, green, blue) float values (between 0.0 and
        1.0).
        """
        return (self.red, self.green, self.blue)

    @property
    def rgb_565(self):
        """
        Returns an unsigned 16-bit integer number representing the color in
        the RGB565 encoding.
        """
        return (
                (int(self.red   * 0xF800) & 0xF800) |
                (int(self.green * 0x07E0) & 0x07E0) |
                (int(self.blue  * 0x001F) & 0x001F))

    @property
    def rgb_bytes(self):
        """
        Returns a 3-tuple of (red, green, blue) byte values.
        """
        return (
                int(self.red * 255),
                int(self.green * 255),
                int(self.blue * 255),
                )

    @property
    def yuv(self):
        """
        Returns a 3-tuple of (y, u, v) float values; y values can be between
        0.0 and 1.0, u values are between -0.436 and 0.436, and v values are
        between -0.615 and 0.615.
        """
        r, g, b = self.rgb
        y = 0.299 * r + 0.587 * g + 0.114 * b
        return (
                y,
                0.492 * (b - y),
                0.877 * (r - y),
                )

    @property
    def yuv_bytes(self):
        """
        Returns a 3-tuple of (y, u, v) byte values. Y values are biased by 16
        in the result to prevent negatives. U and V values are biased by 128
        for the same purpose.
        """
        r, g, b = self.rgb_bytes
        return (
                (( 66 * r + 129 * g +  25 * b + 128) >> 8) + 16,
                ((-38 * r -  73 * g + 112 * b + 128) >> 8) + 128,
                ((112 * r -  94 * g -  18 * b + 128) >> 8) + 128,
                )

    @property
    def yiq(self):
        """
        Returns a 3-tuple of (y, i, q) float values; y values can be between
        0.0 and 1.0, whilst i and q values can be between -1.0 and 1.0.
        """
        return colorsys.rgb_to_yiq(self.red, self.green, self.blue)

    @property
    def cie_xyz(self):
        """
        Returns a 3-tuple of (X, Y, Z) float values representing the color in
        the `CIE 1931 color space`_. The conversion assumes the sRGB working
        space, with reference white D65.

        .. _CIE 1931 color space: https://en.wikipedia.org/wiki/CIE_1931_color_space
        """
        return tuple(matrix_mult(
            ((0.4124564, 0.3575761, 0.1804375),
             (0.2126729, 0.7151522, 0.0721750),
             (0.0193339, 0.1191920, 0.9503041),
             ),
            (from_srgb(self.red),
             from_srgb(self.green),
             from_srgb(self.blue)
             )
             ))

    @property
    def cie_lab(self):
        """
        Returns a 3-tuple of (L*, a*, b*) float values representing the color
        in the `CIE Lab color space`_ with the `D65 standard illuminant`_.

        .. _CIE Lab color space: https://en.wikipedia.org/wiki/Lab_color_space
        .. _D65 standard illuminant: https://en.wikipedia.org/wiki/Illuminant_D65
        """
        K = Fraction(1, 3) * Fraction(29, 6) ** 2
        e = Fraction(6, 29) ** 3
        x, y, z = (n / m for n, m in zip(self.cie_xyz, D65))
        fx, fy, fz = (
            n ** Fraction(1, 3) if n > e else K * n + Fraction(4, 29)
            for n in (x, y, z)
            )
        return (116 * fy - 16, 500 * (fx - fy), 200 * (fy - fz))

    @property
    def cie_luv(self):
        """
        Returns a 3-tuple of (L*, u*, v*) float values representing the color
        in the `CIE Luv color space`_ with the `D65 standard illuminant`_.

        .. _CIE Luv color space: https://en.wikipedia.org/wiki/CIELUV
        """
        K = Fraction(29, 3) ** 3
        e = Fraction(6, 29) ** 3
        XYZ = self.cie_xyz
        yr = XYZ[1] / D65[1]
        L = 116 * yr ** Fraction(1, 3) - 16 if yr > e else K * yr
        u = 13 * L * (U(*XYZ) - U(*D65))
        v = 13 * L * (V(*XYZ) - V(*D65))
        return (L, u, v)

    @property
    def hls(self):
        """
        Returns a 3-tuple of (hue, lightness, saturation) float values (between
        0.0 and 1.0).
        """
        return colorsys.rgb_to_hls(self.red, self.green, self.blue)

    @property
    def hsv(self):
        """
        Returns a 3-tuple of (hue, saturation, value) float values (between 0.0
        and 1.0).
        """
        return colorsys.rgb_to_hsv(self.red, self.green, self.blue)

    @property
    def red(self):
        """
        Returns the red component of the color as a :class:`Red` instance which
        can be used in operations with other :class:`Color` instances.
        """
        # super() calls needed here to avoid recursion
        return Red(super(Color, self).red)

    @property
    def green(self):
        """
        Returns the green component of the color as a :class:`Green` instance
        which can be used in operations with other :class:`Color` instances.
        """
        return Green(super(Color, self).green)

    @property
    def blue(self):
        """
        Returns the blue component of the color as a :class:`Blue` instance
        which can be used in operations with other :class:`Color` instances.
        """
        return Blue(super(Color, self).blue)

    @property
    def hue(self):
        """
        Returns the hue of the color as a :class:`Hue` instance which can be
        used in operations with other :class:`Color` instances.
        """
        return Hue(self.hls[0])

    @property
    def lightness(self):
        """
        Returns the lightness of the color as a :class:`Lightness` instance
        which can be used in operations with other :class:`Color` instances.
        """
        return Lightness(self.hls[1])

    @property
    def saturation(self):
        """
        Returns the saturation of the color as a :class:`Saturation` instance
        which can be used in operations with other :class:`Color` instances.
        """
        return Saturation(self.hls[2])

    def difference(self, other, method='euclid'):
        """
        Determines the difference between this color and *other* using the
        specified *method*. The *method* is specified as a string, and the
        following methods are valid:

        * 'euclid' - This is the default method. Calculate the `Euclidian
          distance`_. This is by far the fastest method, but also the least
          accurate in terms of human perception.
        * 'cie1976' - Use the `CIE 1976`_ formula for calculating the
          difference between two colors in CIE Lab space.
        * 'cie1994g' - Use the `CIE 1994`_ formula with the "graphic arts" bias
          for calculating the difference.
        * 'cie1994t' - Use the `CIE 1994`_ forumula with the "textiles" bias
          for calculating the difference.
        * 'cie2000' - Use the `CIEDE 2000`_ formula for calculating the
          difference.

        Note that the Euclidian distance will be significantly different to the
        other calculations; effectively this just measures the distance between
        the two colors by treating them as coordinates in a three dimensional
        Euclidian space. All other methods are means of calculating a `Delta
        E`_ value in which 2.3 is considered a `just-noticeable difference`_
        (JND).

        .. warning::

            This implementation has yet to receive any significant testing
            (constructor methods for CIELab need to be added before this can be
            done).

        .. _Delta E: https://en.wikipedia.org/wiki/Color_difference
        .. _just-noticeable difference: https://en.wikipedia.org/wiki/Just-noticeable_difference
        .. _Euclidian distance: https://en.wikipedia.org/wiki/Euclidean_distance
        .. _CIE 1976: https://en.wikipedia.org/wiki/Color_difference#CIE76
        .. _CIE 1994: https://en.wikipedia.org/wiki/Color_difference#CIE94
        .. _CIEDE 2000: https://en.wikipedia.org/wiki/Color_difference#CIEDE2000
        """
        if isinstance(method, bytes):
            method = method.decode('ascii')
        if method == 'euclid':
            return sqrt(sum((Cs - Co) ** 2 for Cs, Co in zip(self, other)))
        elif method == 'cie1976':
            return self._cie1976(other)
        elif method.startswith('cie1994'):
            return self._cie1994(other, method)
        elif method == 'ciede2000':
            return self._ciede2000(other)
        else:
            raise ValueError('invalid method: %s' % method)

    def _cie1976(self, other):
        return sqrt(sum((Cs - Co) ** 2 for Cs, Co in zip(self.cie_lab, other.cie_lab)))

    def _cie1994(self, other, method):
        L1, a1, b1 = self.cie_lab
        L2, a2, b2 = other.cie_lab
        dL = L1 - L2
        C1 = sqrt(a1 ** 2 + b1 ** 2)
        C2 = sqrt(a2 ** 2 + b2 ** 2)
        dC = C1 - C2
        # Don't bother with the sqrt here as due to limited float precision
        # we can wind up with a domain error (because the value is ever so
        # slightly negative - try it with black'n'white for an example)
        dH2 = (a1 - a2) ** 2 + (b1 - b2) ** 2 - dC ** 2
        kL, K1, K2 = {
            'cie1994g': (1, 0.045, 0.015),
            'cie1994t': (2, 0.048, 0.014),
            }[method]
        SC = 1 + K1 * C1
        SH = 1 + K2 * C1
        return sqrt((dL ** 2 / kL) + (dC ** 2 / SC) + (dH2 / SH))

    def _ciede2000(self, other):
        L1, a1, b1 = self.cie_lab
        L2, a2, b2 = other.cie_lab
        L_ = (L1 + L2) / 2
        dL = L2 - L1
        C1 = sqrt(a1 ** 2 + b1 ** 2)
        C2 = sqrt(a1 ** 2 + b1 ** 2)
        C_ = (C1 + C2) / 2
        dC = C2 - C1
        G = (1 - sqrt(C_ ** 7 / (C_ ** 7 + 25 ** 7))) / 2
        a1 = (1 + G) * a1
        a2 = (1 + G) * a2
        h1 = 0.0 if b1 == a1 == 0 else degrees(atan2(b1, a1)) % 360
        h2 = 0.0 if b2 == a2 == 0 else degrees(atan2(b2, a2)) % 360
        if C1 * C2 == 0.0:
            dh = 0.0
            h_ = h1 + h2
        elif abs(h1 - h2) <= 180:
            dh = h2 - h1
            h_ = (h1 + h2) / 2
        else:
            if h2 <= h1:
                dh = h2 - h1 + 360
            else:
                dh = h2 - h1 - 360
            if h1 + h2 >= 360:
                h_ = (h1 + h2 + 360) / 2
            else:
                h_ = (h1 + h2 - 360) / 2
        dH = 2 * sqrt(C1 * C2) * sin(radians(dh / 2))
        T = (
                1 -
                0.17 * cos(radians(h_ - 30)) +
                0.24 * cos(radians(2 * h_)) +
                0.32 * cos(radians(3 * h_ + 6)) -
                0.20 * cos(radians(4 * h_ - 63))
                )
        SL = 1 + (0.015 * (L_ - 50) ** 2) / sqrt(20 + (L_ - 50) ** 2)
        SC = 1 + 0.045 * C_
        SH = 1 + 0.015 * C_ * T
        RT = -2 * sqrt(C_ ** 7 / (C_ ** 7 + 25 ** 7)) * sin(radians(60 * exp(-(((h_ - 275) / 25) ** 2))))
        return sqrt((dL / SL) ** 2 + (dC / SC) ** 2 + (dH / SH) ** 2 + RT * (dC / SC) * (dH / SH))




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Python camera library for the Rasperry-Pi camera module
# Copyright (c) 2013-2015 Dave Jones <dave@waveform.org.uk>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import (
    unicode_literals,
    print_function,
    division,
    absolute_import,
    )

# Make Py2's str equivalent to Py3's
str = type('')

import mimetypes
import ctypes as ct
from functools import reduce
from operator import or_

from . import bcm_host, mmalobj as mo, mmal
from .encoders import PiCookedOneImageEncoder, PiRawOneImageEncoder
from .exc import PiCameraRuntimeError, PiCameraValueError


class PiDisplay(object):
    __slots__ = (
        '_display',
        '_info',
        '_transform',
        '_exif_tags',
        )

    _ROTATIONS = {
        bcm_host.DISPMANX_NO_ROTATE:  0,
        bcm_host.DISPMANX_ROTATE_90:  90,
        bcm_host.DISPMANX_ROTATE_180: 180,
        bcm_host.DISPMANX_ROTATE_270: 270,
        }
    _ROTATIONS_R = {v: k for k, v in _ROTATIONS.items()}
    _ROTATIONS_MASK = reduce(or_, _ROTATIONS.keys(), 0)

    RAW_FORMATS = {
        'yuv',
        'rgb',
        'rgba',
        'bgr',
        'bgra',
        }

    def __init__(self, display_num=0):
        bcm_host.bcm_host_init()
        self._exif_tags = {}
        self._display = bcm_host.vc_dispmanx_display_open(display_num)
        self._transform = bcm_host.DISPMANX_NO_ROTATE
        if not self._display:
            raise PiCameraRuntimeError('unable to open display %d' % display_num)
        self._info = bcm_host.DISPMANX_MODEINFO_T()
        if bcm_host.vc_dispmanx_display_get_info(self._display, self._info):
            raise PiCameraRuntimeError('unable to get display info')

    def close(self):
        bcm_host.vc_dispmanx_display_close(self._display)
        self._display = None

    @property
    def closed(self):
        return self._display is None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def _get_output_format(self, output):
        """
        Given an output object, attempt to determine the requested format.

        We attempt to determine the filename of the *output* object and derive
        a MIME type from the extension. If *output* has no filename, an error
        is raised.
        """
        if isinstance(output, bytes):
            filename = output.decode('utf-8')
        elif isinstance(output, str):
            filename = output
        else:
            try:
                filename = output.name
            except AttributeError:
                raise PiCameraValueError(
                    'Format must be specified when output has no filename')
        (type, encoding) = mimetypes.guess_type(filename, strict=False)
        if not type:
            raise PiCameraValueError(
                'Unable to determine type from filename %s' % filename)
        return type

    def _get_image_format(self, output, format=None):
        """
        Given an output object and an optional format, attempt to determine the
        requested image format.

        This method is used by all capture methods to determine the requested
        output format. If *format* is specified as a MIME-type the "image/"
        prefix is stripped. If *format* is not specified, then
        :meth:`_get_output_format` will be called to attempt to determine
        format from the *output* object.
        """
        if isinstance(format, bytes):
            format = format.decode('utf-8')
        format = format or self._get_output_format(output)
        format = (
            format[6:] if format.startswith('image/') else
            format)
        if format == 'x-ms-bmp':
            format = 'bmp'
        return format

    def _get_image_encoder(self, output_port, format, resize, **options):
        """
        Construct an image encoder for the requested parameters.

        This method is called by :meth:`capture`. The *output_port* parameter
        gives the MMAL port that the encoder should read output from. The
        *format* parameter indicates the image format and will be one of:

        * ``'jpeg'``
        * ``'png'``
        * ``'gif'``
        * ``'bmp'``
        * ``'yuv'``
        * ``'rgb'``
        * ``'rgba'``
        * ``'bgr'``
        * ``'bgra'``

        The *resize* parameter indicates the size that the encoder should
        resize the output to (presumably by including a resizer in the
        pipeline). Finally, *options* includes extra keyword arguments that
        should be passed verbatim to the encoder.
        """
        encoder_class = (
                PiRawOneImageEncoder if format in self.RAW_FORMATS else
                PiCookedOneImageEncoder)
        return encoder_class(
                self, None, output_port, format, resize, **options)

    def capture(self, output, format=None, resize=None, **options):
        format = self._get_image_format(output, format)
        if format == 'yuv':
            raise PiCameraValueError('YUV format is unsupported at this time')
        res = self.resolution
        if (self._info.transform & bcm_host.DISPMANX_ROTATE_90) or (
                self._info.transform & bcm_host.DISPMANX_ROTATE_270):
            res = res.transpose()
        transform = self._transform
        if (transform & bcm_host.DISPMANX_ROTATE_90) or (
                transform & bcm_host.DISPMANX_ROTATE_270):
            res = res.transpose()
        source = mo.MMALPythonSource()
        source.outputs[0].format = mmal.MMAL_ENCODING_RGB24
        if format == 'bgr':
            source.outputs[0].format = mmal.MMAL_ENCODING_BGR24
            transform |= bcm_host.DISPMANX_SNAPSHOT_SWAP_RED_BLUE
        source.outputs[0].framesize = res
        source.outputs[0].commit()
        encoder = self._get_image_encoder(
            source.outputs[0], format, resize, **options)
        try:
            encoder.start(output)
            try:
                pitch = res.pad(width=16).width * 3
                image_ptr = ct.c_uint32()
                resource = bcm_host.vc_dispmanx_resource_create(
                    bcm_host.VC_IMAGE_RGB888, res.width, res.height, image_ptr)
                if not resource:
                    raise PiCameraRuntimeError(
                        'unable to allocate resource for capture')
                try:
                    buf = source.outputs[0].get_buffer()
                    if bcm_host.vc_dispmanx_snapshot(self._display, resource, transform):
                        raise PiCameraRuntimeError('failed to capture snapshot')
                    rect = bcm_host.VC_RECT_T(0, 0, res.width, res.height)
                    if bcm_host.vc_dispmanx_resource_read_data(resource, rect, buf._buf[0].data, pitch):
                        raise PiCameraRuntimeError('failed to read snapshot')
                    buf._buf[0].length = pitch * res.height
                    buf._buf[0].flags = (
                        mmal.MMAL_BUFFER_HEADER_FLAG_EOS |
                        mmal.MMAL_BUFFER_HEADER_FLAG_FRAME_END
                        )
                finally:
                    bcm_host.vc_dispmanx_resource_delete(resource)
                source.outputs[0].send_buffer(buf)
                # XXX Anything more intelligent than a 10 second default?
                encoder.wait(10)
            finally:
                encoder.stop()
        finally:
            encoder.close()

    def _calculate_transform(self):
        """
        Calculates a reverse transform to undo any that the boot configuration
        applies (presumably the user has altered the boot configuration to
        match their screen orientation so they want any capture to appear
        correctly oriented by default). This is then modified by the transforms
        specified in the :attr:`rotation`, :attr:`hflip` and :attr:`vflip`
        attributes.
        """
        r = PiDisplay._ROTATIONS[self._info.transform & PiDisplay._ROTATIONS_MASK]
        r = (360 - r) % 360 # undo the native rotation
        r = (r + self.rotation) % 360 # add selected rotation
        result = PiDisplay._ROTATIONS_R[r]
        result |= self._info.transform & ( # undo flips by re-doing them
            bcm_host.DISPMANX_FLIP_HRIZ | bcm_host.DISPMANX_FLIP_VERT
            )
        return result

    @property
    def resolution(self):
        """
        Retrieves the resolution of the display device.
        """
        return mo.PiResolution(width=self._info.width, height=self._info.height)

    def _get_hflip(self):
        return bool(self._info.transform & bcm_host.DISPMANX_FLIP_HRIZ)
    def _set_hflip(self, value):
        if value:
            self._info.transform |= bcm_host.DISPMANX_FLIP_HRIZ
        else:
            self._info.transform &= ~bcm_host.DISPMANX_FLIP_HRIZ
    hflip = property(_get_hflip, _set_hflip, doc="""\
        Retrieves or sets whether snapshots are horizontally flipped.

        When queried, the :attr:`vflip` property returns a boolean indicating
        whether or not the output of :meth:`capture` is horizontally flipped.
        The default is ``False``.

        .. note::

            This property only affects snapshots; it does not affect the
            display output itself.
        """)

    def _get_vflip(self):
        return bool(self._info.transform & bcm_host.DISPMANX_FLIP_VERT)
    def _set_vflip(self, value):
        if value:
            self._info.transform |= bcm_host.DISPMANX_FLIP_VERT
        else:
            self._info.transform &= ~bcm_host.DISPMANX_FLIP_VERT
    vflip = property(_get_vflip, _set_vflip, doc="""\
        Retrieves or sets whether snapshots are vertically flipped.

        When queried, the :attr:`vflip` property returns a boolean indicating
        whether or not the output of :meth:`capture` is vertically flipped. The
        default is ``False``.

        .. note::

            This property only affects snapshots; it does not affect the
            display output itself.
        """)

    def _get_rotation(self):
        return PiDisplay._ROTATIONS[self._transform & PiDisplay._ROTATIONS_MASK]
    def _set_rotation(self, value):
        try:
            self._transform = (
                self._transform & ~PiDisplay._ROTATIONS_MASK) | PiDisplay._ROTATIONS_R[value]
        except KeyError:
            raise PiCameraValueError('invalid rotation %d' % value)
    rotation = property(_get_rotation, _set_rotation, doc="""\
        Retrieves or sets the rotation of snapshots.

        When queried, the :attr:`rotation` property returns the rotation
        applied to the result of :meth:`capture`. Valid values are 0, 90, 180,
        and 270. When set, the property changes the rotation applied to the
        result of :meth:`capture`. The default is 0.

        .. note::

            This property only affects snapshots; it does not affect the
            display itself. To rotate the display itself, modify the
            ``display_rotate`` value in :file:`/boot/config.txt`.
        """)

    def _get_exif_tags(self):
        return self._exif_tags
    def _set_exif_tags(self, value):
        self._exif_tags = {k: v for k, v in value.items()}
    exif_tags = property(_get_exif_tags, _set_exif_tags)

